/*----------------------------------------------------------------------------
* Testprogramm fuer ubasic_avr-Interpreter....
* 
* Uwe Berger (bergeruw@gmx.net); 2010
* 
* 
* Ein-/Ausgaben erfolgen ueber die serielle Schnittstelle (57600, 8, n, 1)
*
* Kommandos:
*        list   --> gibt das aktuelle Basic-Programm aus
*        run    --> startet das aktuell geladene Basic-Programm; das
*                   laufende Programm kann man mit einem [ESC] abbrechen
*        load   --> die folgenden Eingaben werden als Basic-Programm gewertet
*                   mit einem [ESC] beendet man den Eingabe-Mode
*        load n --> laed das Programm mit der angegenen Nummer n (siehe
*                   auch ubasic_tests.h)
*        mem    --> Infos ueber RAM (unused, free)
*
* Have fun!
* ---------
* 
* ----------------------------------------------------------------------------*/

#ifdef AVR
	#include <avr/io.h>
	#include "usart.h"
#endif

#ifdef WIN32
	#define PROGMEM 
	#include <stdio.h>
	#include <string.h>
	#include "platform.h"

	unsigned char usart_receive_char (void);
	unsigned char usart_is_receive(void);
	unsigned char usart_read_line(char* buffer, unsigned char  buffer_length);
	int get_mem_unused(void);
	int get_mem_free(void);

#endif

#include "ubasic.h"
#include "ubasic_tests.h"
#include "mem_check.h"

#include "tokenizer.h"

#include <stdio.h>

//Taktfrequenz
#ifndef F_CPU
#define F_CPU 16000000UL
#endif

//Baudrate der seriellen Schnittstelle
#define BAUDRATE 38400



#define MAX_PROGRAM_LEN	400

static char program[MAX_PROGRAM_LEN] = "";

//----------------------------------------------------------------------------
void basic_load( char* Buffer){
	int i = 0;
	char c;

	while((c=usart_receive_char()) != 0x1b && i < MAX_PROGRAM_LEN - 1) {
		if (c=='\r' || c=='\n') {
			c='\n';
			printf_P(PSTR ("\r\n"));
		} else {
			#ifdef AVR
				usart_write_char(c);
			#endif
			#ifdef WIN32
				printf_P (PSTR("%c"), c);
			#endif
		}
		Buffer[i++]=c;
	}
	Buffer[i] = '\0';
	printf_P(PSTR ("\r\nEnde Eingabemodus...\r\n"));
}

//----------------------------------------------------------------------------
void basic_list(char *c) {
	while (*c) {
		#ifdef AVR
			if (*c == '\n') usart_write_char('\r');
			usart_write_char(*c++);
		#endif
		#ifdef WIN32
			if (*c == '\n') printf_P(PSTR ("\r"));
			printf_P (PSTR("%c"), *c++);
		#endif
	}
}

//----------------------------------------------------------------------------
void load_from_flash(char* Buffer, int prog_idx) {
	const char *pstrflash;
	if (prog_idx >= (sizeof(progs)/sizeof(progs[0]))) {
		printf_P(PSTR ("\r\nProgramm existiert nicht!\r\n"));
	} else {
		pstrflash = (const char*)(pgm_read_ptr(&(progs[prog_idx])));
		memcpy_P(Buffer, pstrflash, MAX_PROGRAM_LEN);
		printf_P(PSTR ("\n\rProgramm %i geladen.\n\r"), prog_idx);
	}
}



#ifdef WIN32
	unsigned char usart_receive_char (void)
	{
		char x = getch();
		if (x == 0)			{ getch(); return 0x20; }
		if (x == 0xe0)	{ getch(); return 0x20; }
		return x;
	}

	unsigned char usart_is_receive(void)
	{
		return kbhit();
	}

	unsigned char usart_read_line(char* buffer, unsigned char  buffer_length)
	{
			unsigned char read_length = 0;

			memset(buffer, 0, buffer_length);
			while(read_length < buffer_length - 1)
			{
					unsigned char  c = usart_receive_char();
					if(c == 0x08 || c == 0x7f)
					{
							if(read_length < 1) continue;
							--read_length;
							buffer[read_length] = '\0';
							printf("\x08 \x08");
							continue;
					}
					printf("%c",c);
					if(c == 0x0d)
					{
							buffer[read_length] = '\0';
							break;
					}
					else
					{
							buffer[read_length] = c;
							++read_length;
					}
			}
			return read_length;
	}

	int get_mem_unused(void)
	{
		return 0;
	}

	int get_mem_free(void)
	{
		return 0;
	}

#endif





int systick = 12287;

int adc [8] = { 1, 2, 3, 4, 5, 6, 7, 8, };



int lcdoutfunc (T_USER_FUNC_DATAS stDatas)
{
	printf ("lcd write %d %d %s\n\r", stDatas [0].value, stDatas [1].value, stDatas [2].text);
	return 0;
}


int halffunc (T_USER_FUNC_DATAS stDatas)
{
	if (stDatas [0].type != 1) { FATAL_ABORT (); }
	return stDatas [0].value / 2;
}

int dblfunc (T_USER_FUNC_DATAS stDatas)
{
	if (stDatas [0].type != 1) { FATAL_ABORT (); }
	return stDatas [0].value * 2;
}

int resetfunc (T_USER_FUNC_DATAS stDatas)
{
	return 0;
}

#ifdef AVR

	int uart1file_putchar (char x, FILE *stream);

	FILE uart1file 		= FDEV_SETUP_STREAM (uart1file_putchar, NULL, _FDEV_SETUP_WRITE);

	int uart1file_putchar (char x, FILE *stream)
	{
		usart_write_char (x);
		return 0;
	}
#endif

//----------------------------------------------------------------------------
//Hier startet das Hauptprogramm
int main(void)
{  
	char buffer[10];


#ifdef AVR
	stdout = &uart1file;
	
  usart_init(BAUDRATE); // setup the UART
#endif

	
	printf_P(PSTR ("AVR-uBasic-Interpreter; Uwe Berger, 2010\r\n"));
	printf_P(PSTR ("Compiliert am "__DATE__" um "__TIME__"\r\n"));
#ifdef AVR
	printf_P(PSTR ("Compiliert mit GCC Version "__VERSION__"\r\n"));
#endif
	// Prompt
  printf_P(PSTR ("\r\n>"));
	
	while(1){

		char* command = buffer;
        if(usart_read_line(command, sizeof(buffer)) < 1) continue;
        // Basic-Programm starten
        if (strcmp_P(command, PSTR("run")) == 0) {
			printf_P(PSTR ("\r\n"));
			ubasic_init(program);
			do {
				ubasic_run();
				// Basic-Programm unterbrechen, wenn [ESC] empfangen
				if (usart_is_receive() && (usart_receive_char() == 0x1b)) {
					printf_P(PSTR ("\r\nProgramm unterbrochen...!\r\n"));
					break;
				}
			} while(!ubasic_finished());
        } else
        // Programm aus Programmflash laden
        if (strncmp_P(command, PSTR("load "), 5) == 0) {
			command += 5;
			if(command[0] == '\0') continue;
			load_from_flash(program, atoi(command));
        } else
        // Programmeingabe via serieller Schnittstelle
        if (strcmp_P(command, PSTR("load")) == 0) {
			printf_P(PSTR ("\r\nBeginn manueller Eingabemodus, Ende mit [ESC].\r\n"));
			basic_load(program);			
        } else
        // ein paar Infos ueber den RAM
        if (strcmp_P(command, PSTR("mem")) == 0) {
			printf_P(PSTR ("\r\nunused mem: %i\r\n"), get_mem_unused());
			printf_P(PSTR ("free mem..: %i\r\n"), get_mem_free());
        } else
        // Basic-Programm auflisten
        if (strcmp_P(command, PSTR("list")) == 0) {
			printf_P(PSTR ("\r\n"));
			basic_list(program);
        } else 	printf_P(PSTR ("\r\nKommando %s unbekannt!\r\n"), command);
        // Prompt
        printf_P(PSTR ("\r\n>"));
    }
		
return(0);
}

