/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * ---------------------------------------------------------------------------
 * Source modified by Uwe Berger (bergeruw@gmx.net); 2010
 * Additional modification by Rene Bllhoff (reneboellhoff@gmx.de); Jun 2010
 * ---------------------------------------------------------------------------
 */

//#define DEBUG 0

#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif

#include "platform.h"

#include "tokenizer.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

char *ptr, *nextptr;

#define MAX_NUMLEN 5

typedef struct 
{
  char *keyword;
  int token;
	T_BASIC_FUNCTION func;
} T_KEYWORD_TOKEN;

int current_token = TOKENIZER_ERROR;

char userkeyword [MAX_USER_KEYWORD_LEN];

int 	userkeywordindex 	= 0;
int 	userkeywordtype  	= 0;
int 	userkeyworddata  	= 0;
void *userkeywordptr 		= NULL;


#define BASIC_TOKENS

	#define TOKEN_WORD(name,text,func) char keyword__##name [] FLASHMEM = text;
	#define TOKEN_CHAR(name,char,func)
		#include BASIC_CONFIG_FILE
	#undef TOKEN_WORD
	#undef TOKEN_CHAR


	T_KEYWORD_TOKEN keywords[] FLASHMEM = {
		#define TOKEN_WORD(name,text,func) { keyword__##name, name, func, },
		#define TOKEN_CHAR(name,char,func)
			#include BASIC_CONFIG_FILE
		#undef TOKEN_WORD
		#undef TOKEN_CHAR
		{NULL, TOKENIZER_ERROR}
	};


	T_BASIC_EXPR_FUNC basicopfuncs [] FLASHMEM = 
	{
		#define TOKEN_WORD(name,text,func)
		#define TOKEN_CHAR(name,char,func) func,
			#include BASIC_CONFIG_FILE
		#undef TOKEN_WORD
		#undef TOKEN_CHAR
	};


#undef BASIC_TOKENS

#define BASIC_USER_EXTENSIONS

		
		#define USER_FUNCTION(name,text,func,param)	char userfunc__##name [] FLASHMEM  = { text };
		#define USER_VARIABLE(name,text,var)				char uservar__##name  [] FLASHMEM  = { text };
		#define USER_ARRAY(name,text,var,size)			char userarr__##name  [] FLASHMEM  = { text };	
			#include BASIC_CONFIG_FILE
		#undef USER_FUNCTION
		#undef USER_VARIABLE
		#undef USER_ARRAY

		char *userkeywords [] FLASHMEM  = {
			#define USER_FUNCTION(name,text,func,param)	userfunc__##name,
			#define USER_VARIABLE(name,text,var)				uservar__##name, 
			#define USER_ARRAY(name,text,var,size)			userarr__##name, 	
				#include BASIC_CONFIG_FILE
			#undef USER_FUNCTION
			#undef USER_VARIABLE
			#undef USER_ARRAY
		};

		void *userpointer [] FLASHMEM  = {
			#define USER_FUNCTION(name,text,func,param)	(void *) func,	
			#define USER_VARIABLE(name,text,var)				(void *) &var,
			#define USER_ARRAY(name,text,var,size)			(void *) &var,
				#include BASIC_CONFIG_FILE
			#undef USER_FUNCTION
			#undef USER_VARIABLE
			#undef USER_ARRAY
		};

		char usertypes [] FLASHMEM  = {
			#define USER_FUNCTION(name,text,func,param)	1,
			#define USER_VARIABLE(name,text,var)				2,
			#define USER_ARRAY(name,text,var,size)			3,
				#include BASIC_CONFIG_FILE
			#undef USER_FUNCTION
			#undef USER_VARIABLE
			#undef USER_ARRAY
		};

		int userdatas [] FLASHMEM  = {
			#define USER_FUNCTION(name,text,func,param)	param,
			#define USER_VARIABLE(name,text,var)				0,
			#define USER_ARRAY(name,text,var,size)			size,
				#include BASIC_CONFIG_FILE
			#undef USER_FUNCTION
			#undef USER_VARIABLE
			#undef USER_ARRAY
		};

#undef BASIC_USER_EXTENSIONS


/*---------------------------------------------------------------------------*/

int singlechar(void)
{
	#define BASIC_TOKENS
																				if (0) { }
		#define TOKEN_WORD(name,text,func)
		#define TOKEN_CHAR(name,char,func)  else if (*ptr == char) 	return name;
			#include BASIC_CONFIG_FILE
		#undef TOKEN_CHAR
		#undef TOKEN_WORD
																				else 										return 0;
	#undef BASIC_TOKENS
}

void execbasicfunc (void)
{
	unsigned char temp;
	T_BASIC_FUNCTION func;
	if ((current_token >= TOKENIZER_KEYWORDS) && (current_token < TOKENIZER_KEYCHARS))
	{
	  temp = current_token - TOKENIZER_KEYWORDS - 1;
	  func = (T_BASIC_FUNCTION) pgm_read_ptr (&(keywords [temp].func));
	  func();
	}
}

int execbasicexprfunc (int op, int a, int b)
{
  T_BASIC_EXPR_FUNC func;
	if ((op >= TOKENIZER_KEYCHARS) && (op < TOKENIZER_NUM_TOKENS))
	{
		op = op - TOKENIZER_KEYCHARS;
		func = (T_BASIC_EXPR_FUNC) pgm_read_ptr (&(basicopfuncs [op]));
		return func (a, b);

	}
	return 0;
}

int searchuserkeyword (void)
{
	unsigned char c;
	char *keyword;

	for (c = 0; c < BASIC_NUM_USER_EXTENSIONS; c++)
	{
		keyword = pgm_read_ptr (&(userkeywords [c]));
		if (strcmp_P (userkeyword, keyword) == 0)
		{
			userkeywordindex 	= c;
			userkeyworddata 	= pgm_read_byte (&(userdatas 	[c]));
			userkeywordtype 	= pgm_read_byte (&(usertypes 	[c]));
			userkeywordptr  	= pgm_read_ptr  (&(userpointer [c]));
			return c;
		}
	}
	userkeywordindex 	= -1;
	userkeyworddata 	= 0;
	userkeywordtype 	= 0;
	userkeywordptr 		= NULL;
	return -1;
}

int tokenizer_userdata (int *type, int *data, void **ptr)
{
	*type = userkeywordtype;
	*data = userkeyworddata;
	*ptr  = userkeywordptr;
	return userkeywordindex;
}


/*---------------------------------------------------------------------------*/
int get_next_token(void)
{
	char *keyword;
  int i;
	
  DEBUG_PRINTF("get_next_token(): '%s'\n", ptr);

  if(*ptr == 0) {
    return TOKENIZER_ENDOFINPUT;
  }
  
  if(isdigit(*ptr)) {
    for(i = 0; i < MAX_NUMLEN; ++i) {
      if(!isdigit(ptr[i])) {
				if(i > 0) {
					nextptr = ptr + i;
					return TOKENIZER_NUMBER;
				} else {
					DEBUG_PRINTF("get_next_token: error due to too short number\n");
					return TOKENIZER_ERROR;
				}
						}
						if(!isdigit(ptr[i])) {
				DEBUG_PRINTF("get_next_token: error due to malformed number\n");
				return TOKENIZER_ERROR;
      }
    }
    DEBUG_PRINTF("get_next_token: error due to too long number\n");
    return TOKENIZER_ERROR;
  } else if(singlechar()) {
    nextptr = ptr + 1;
    return singlechar();
  } else if(*ptr == '"') {
    nextptr = ptr;
    do {
      ++nextptr;
    } while(*nextptr != '"');
    ++nextptr;
    return TOKENIZER_STRING;
  } else {
		
    for(i = 0; i < (sizeof (keywords)/sizeof (keywords[0]));i++) {
		  keyword = pgm_read_ptr (&(keywords [i].keyword));
			if (keyword == NULL) break;
      if(strncmp_P(ptr, keyword, strlen_P(keyword)) == 0) {
				nextptr = ptr + strlen_P(keyword);
				return pgm_read_word (&(keywords [i].token));
      }
    }
	}

  if (*ptr >= 'a' && *ptr <= 'z') {
		if (!((ptr [1] >= 'a') && (ptr [1] <= 'z')))
		{
			nextptr = ptr + 1;
			return TOKENIZER_VARIABLE;
		} else {
			memset (userkeyword, 0, sizeof (userkeyword));
			for (i = 0; i < MAX_USER_KEYWORD_LEN - 1; i++)
			{
				if (i == 0) 
				{
					if (isalpha (*ptr) || (*ptr == '_'))  { userkeyword [i] = *ptr; ptr++; }
																								else { break; }
				}
				else if (i != 0)
				{
					if (isalpha (*ptr) || (*ptr == '_') || isdigit(*ptr))  { userkeyword [i] = *ptr;  ptr++; }
																								else { break; }
				}
			}
			if (i == 0)  { return TOKENIZER_ERROR; }
							else { nextptr = ptr;}
			searchuserkeyword ();
			return TOKENIZER_NAME;
		}
	}
  return TOKENIZER_ERROR;
}
/*---------------------------------------------------------------------------*/
void tokenizer_init(const char *program)
{
  ptr = program;
  current_token = get_next_token();
}
/*---------------------------------------------------------------------------*/
int tokenizer_token(void)
{
  return current_token;
}
/*---------------------------------------------------------------------------*/
void tokenizer_next(void)
{

  if(tokenizer_finished()) {
    return;
  }

  DEBUG_PRINTF("tokenizer_next: %p\n", nextptr);
  ptr = nextptr;
  while(*ptr == ' ') {
    ++ptr;
  }
  current_token = get_next_token();
  DEBUG_PRINTF("tokenizer_next: '%s' %d\n", ptr, current_token);
  return;
}
/*---------------------------------------------------------------------------*/
int tokenizer_num(void)
{
  return atoi(ptr);
}
/*---------------------------------------------------------------------------*/
void tokenizer_string(char *dest, int len)
{
  char *string_end;
  int string_len;
  
  if(tokenizer_token() != TOKENIZER_STRING) {
    return;
  }
  string_end = strchr(ptr + 1, '"');
  if(string_end == NULL) {
    return;
  }
  string_len = string_end - ptr - 1;
  if(len < string_len) {
    string_len = len;
  }
  memcpy(dest, ptr + 1, string_len);
  dest[string_len] = 0;
}
/*---------------------------------------------------------------------------*/
void tokenizer_error_print(void)
{
  DEBUG_PRINTF("tokenizer_error_print: '%s'\n", ptr);
}
/*---------------------------------------------------------------------------*/
int tokenizer_finished(void)
{
  return *ptr == 0 || current_token == TOKENIZER_ENDOFINPUT;
}
/*---------------------------------------------------------------------------*/
int tokenizer_variable_num(void)
{
  return *ptr - 'a' + 1;
}
/*---------------------------------------------------------------------------*/
